import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)

import os

import dask.dataframe as dd
import geopandas
import igraph as ig
import leidenalg as la
import numpy as np
import pandas as pd
import rasterio
import torch
import torchvision
from joblib import Parallel, delayed
from rasterio import features
from scipy.stats import zscore
from sklearn.neighbors import NearestNeighbors
from tqdm.auto import tqdm
from umap import UMAP

from ..preprocessing import get_points


def gene_leiden(data, copy=False):
    adata = data.copy() if copy else data

    coloc_sim = (
        adata.uns["coloc_sim_agg"]
        .pivot_table(index="g1", columns="g2", values="coloc_sim")
        .fillna(0)
    )
    coloc_sim = coloc_sim.dropna()

    genes = coloc_sim.index

    # Z scale features
    coloc_sim = zscore(coloc_sim, axis=0)

    nn = NearestNeighbors().fit(coloc_sim)
    connectivity = nn.kneighbors_graph(coloc_sim, n_neighbors=5).toarray()

    loc_umap = UMAP().fit_transform(connectivity)
    loc_umap = pd.DataFrame(loc_umap, index=genes)
    adata.varm["loc_umap"] = loc_umap.reindex(adata.var_names)

    return adata if copy else None


def coloc_cluster_genes(data, resolution=1, copy=False):
    adata = data.copy() if copy else data

    coloc_sim = (
        adata.uns["coloc_sim_agg"]
        .pivot_table(index="g1", columns="g2", values="coloc_sim")
        .fillna(0)
    )

    genes = coloc_sim.index.tolist()
    coloc_sim = coloc_sim.values

    # Z scale features
    coloc_sim = zscore(coloc_sim, axis=0)

    nn = NearestNeighbors().fit(coloc_sim)
    connectivity = nn.kneighbors_graph(coloc_sim, n_neighbors=5).toarray()

    g = ig.Graph.Adjacency(connectivity)
    g.es["weight"] = connectivity[connectivity != 0]
    g.vs["label"] = genes
    partition = la.find_partition(
        g,
        la.CPMVertexPartition,
        weights=g.es["weight"],
        resolution_parameter=resolution,
    )
    gene_clusters = pd.Series(partition.membership, dtype=int, index=g.vs["label"])

    adata.var["coloc_group"] = gene_clusters.reindex(adata.var_names)
    return adata if copy else None


def coloc_sim(data, radius=3, min_count=5, n_cores=1, copy=False):
    """Calculate pairwise gene colocalization similarity using a KNN approach.

    Parameters
    ----------
    adata : AnnData
        Anndata formatted spatial data.
    outer_radius : int, optional
        Number of pixels to search for neighbors, by default 3
    Returns
    -------
    adata : AnnData
        .uns['coloc_sim']: Pairwise gene colocalization similarity within each cell.
    """
    adata = data.copy() if copy else data

    # Filter points and counts by min_count
    counts = adata.to_df()

    # Helper function to apply per cell
    def cell_coloc_sim(p, g_density, name):

        # Get xy coordinates
        xy = p[["x", "y"]].values

        # Get neighbors within fixed outer_radius for every point
        nn = NearestNeighbors(radius=radius).fit(xy)
        distances, point_index = nn.radius_neighbors(xy, return_distance=True)

        # Enumerate point-wise gene labels
        gene_index = p["gene"].reset_index(drop=True)

        # Convert to adjacency list of points, no double counting
        neighbor_pairs = []
        for g1, neighbors, n_dists in zip(gene_index.values, point_index, distances):
            for g2, d in zip(neighbors, n_dists):
                neighbor_pairs.append([g1, g2, d])

        # Calculate pair-wise gene similarity
        neighbor_pairs = pd.DataFrame(neighbor_pairs, columns=["g1", "g2", "p_dist"])

        # Keep minimum distance to g2 point
        neighbor_pairs = neighbor_pairs.groupby(["g1", "g2"]).agg("min").reset_index()
        neighbor_pairs.columns = ["g1", "g2", "point_dist"]

        # Map to gene index
        neighbor_pairs["g2"] = neighbor_pairs["g2"].map(gene_index)

        # Count number of points within distance of increasing radius
        r_step = 0.5
        expected_counts = [
            lambda dists: (dists <= r).sum()
            for r in np.arange(r_step, radius + r_step, r_step)
        ]
        metrics = (
            neighbor_pairs.groupby(["g1", "g2"])
            .agg({"point_dist": expected_counts})
            .reset_index()
        )

        # Colocalization metric: max of L_ij(r) for r <= radius
        g2_density = g_density.loc[metrics["g2"].tolist()].values
        metrics["coloc_sim"] = (
            (metrics["point_dist"].divide(g2_density * np.pi, axis=0))
            .pow(0.5)
            .max(axis=1)
        )
        metrics["cell"] = name

        # Ignore self colocalization
        # metrics = metrics.loc[metrics["g1"] != metrics["g2"]]

        return metrics[["cell", "g1", "g2", "coloc_sim"]]

    # Only keep genes >= min_count in each cell
    gene_densities = []
    counts.apply(lambda row: gene_densities.append(row[row >= min_count]), axis=1)
    # Calculate point density per gene per cell
    gene_densities /= adata.obs["cell_area"]
    gene_densities = gene_densities.values

    cell_metrics = Parallel(n_jobs=n_cores)(
        delayed(cell_coloc_sim)(
            get_points(adata, cells=g_density.name, genes=g_density.index.tolist()),
            g_density,
            g_density.name,
        )
        for g_density in tqdm(gene_densities)
    )

    cell_metrics = pd.concat(cell_metrics)
    cell_metrics.columns = cell_metrics.columns.get_level_values(0)

    # Save coloc similarity
    adata.uns["coloc_sim"] = cell_metrics

    # Aggregate metric across cells
    cell_metrics_agg = dd.from_pandas(cell_metrics, chunksize=1000000)
    agg = (
        cell_metrics_agg.groupby(["g1", "g2"])
        .coloc_sim.sum()
        .compute()
        .reset_index()
    )

    # Mean based on number of cells
    agg['coloc_sim'] /= data.shape[0]

    adata.uns["coloc_sim_agg"] = agg

    return adata if copy else None


# TODO need physical unit size of coordinate system to standardize rendering resolution
def rasterize_cells(
    data,
    imgdir,
    label_layer=None,
    scale_factor=15,
    out_dim=64,
    n_cores=1,
    overwrite=True,
    copy=False,
):
    """Rasterize points and cell masks to grayscale image. Writes directly to file.

    Parameters
    ----------
    data : AnnData
        AnnData formatted spatial data.
    imgdir : str
        Directory where images will be stored.
    """
    adata = data.copy() if copy else data

    os.makedirs(f"{imgdir}", exist_ok=True)

    def write_img(s, n, p, cell_name):

        # Get bounds and size of cell in raw coordinate space
        bounds = s.bounds
        width = bounds[2] - bounds[0]
        height = bounds[3] - bounds[1]

        # Define top left corner for centering/scaling transform
        west = bounds[0] + width / 2 - (out_dim / 2 * scale_factor)
        north = bounds[3] - height / 2 + (out_dim / 2 * scale_factor)

        # Define transform
        tf_origin = rasterio.transform.from_origin(
            west, north, scale_factor, scale_factor
        )

        # Rasterize cell
        base_raster = features.rasterize(
            [s],
            fill=0,
            default_value=20,
            out_shape=(out_dim, out_dim),
            transform=tf_origin,
        )

        # Rasterize nucleus
        if n is not None:
            features.rasterize(
                [n], default_value=40, transform=tf_origin, out=base_raster
            )

        warnings.filterwarnings(
            action="ignore", category=rasterio.errors.NotGeoreferencedWarning
        )

        # Rasterize and write points
        genes = p["gene"].unique().tolist()

        if label_layer:
            labels = dict(
                zip(genes, list(adata[cell_name, genes].layers[label_layer].flatten()))
            )
        else:
            labels = dict(zip(genes, ["foo"] * len(genes)))

        p = geopandas.GeoDataFrame(p, geometry=geopandas.points_from_xy(p["x"], p["y"]))

        for gene_name in genes:
            label = labels[gene_name]

            os.makedirs(f"{imgdir}/{label}", exist_ok=True)

            # TODO implement overwrite param
            if not overwrite and os.path.exists(
                f"{imgdir}/{label}/{cell_name}_{gene_name}.tif"
            ):
                return

            cg_points = p.loc[p["gene"] == gene_name]

            gene_raster = base_raster.copy()

            # Set base as 40
            gene_raster = features.rasterize(
                shapes=cg_points.geometry,
                default_value=40,
                transform=tf_origin,
                out=gene_raster,
            )

            # Plus 20 per point
            features.rasterize(
                shapes=cg_points.geometry,
                default_value=20,
                transform=tf_origin,
                merge_alg=rasterio.enums.MergeAlg("ADD"),
                out=gene_raster,
            )

            # Convert to tensor
            gene_raster = torch.from_numpy(gene_raster.astype(np.float32) / 255)

            torchvision.utils.save_image(
                gene_raster, f"{imgdir}/{label}/{cell_name}_{gene_name}.tif"
            )

    # Parallelize points
    Parallel(n_jobs=n_cores)(
        delayed(write_img)(
            adata.obs.loc[cell_name, "cell_shape"],
            adata.obs.loc[cell_name, "nucleus_shape"],
            get_points(adata, cells=cell_name),
            cell_name,
        )
        for cell_name in tqdm(adata.obs_names.tolist())
    )

    # TODO write filepaths to adata

    return adata if copy else None



import warnings
warnings.filterwarnings("ignore")

import anndata
import geopandas
import numpy as np
import pandas as pd
import scanpy as sc
from anndata import AnnData
from shapely import geometry, wkt


def read_h5ad(filename, backed=None):
    """Load bento processed AnnData object from h5ad.

    Parameters
    ----------
    filename : str
        File name to load data file.
    backed : 'r', 'r+', True, False, None
        If 'r', load AnnData in backed mode instead of fully loading it into memory (memory mode).
        If you want to modify backed attributes of the AnnData object, you need to choose 'r+'.
        By default None.
    Returns
    -------
    AnnData
        AnnData data object.
    """
    adata = anndata.read_h5ad(filename, backed=backed)

    # Load obs columns that are shapely geometries
    adata.obs = adata.obs.apply(
        lambda col: geopandas.GeoSeries(
            col.astype(str).apply(lambda val: wkt.loads(val) if val != "None" else None)
        )
        if col.astype(str).str.startswith("POLYGON").any()
        else geopandas.GeoSeries(col)
    )

    return adata


def write_h5ad(data, filename):
    """Write AnnData to h5ad. Casts each GeoDataFrame in adata.uns['masks'] for h5ad compatibility.

    Parameters
    ----------
    adata : AnnData
        bento loaded AnnData
    filename : str
        File name to write data file.
    """
    # Convert geometry from GeoSeries to list for h5ad serialization compatibility
    adata = data.copy()

    adata.obs = adata.obs.apply(
        lambda col: col.apply(lambda val: val.wkt if val is not None else val).astype(
            str
        )
        if col.astype(str).str.startswith("POLYGON").any()
        else col
    )

    # Write to h5ad
    adata.write(filename)


def read_geodata(points, cell, other={}):
    """Load spots and masks for many cells.

    Parameters
    ----------
    points : str
        Filepath to spots .shp file. Expects GeoDataFrame with geometry of Points, and 'gene' column at minimum.
    cell : str
        Filepath to cell segmentation masks .shp file. Expects GeoDataFrame with geometry of Polygons.
    other : dict(str)
        Filepaths to all other segmentation masks .shp files; expects GeoDataFrames of same format.
        Use keys of dict to access corresponding outputs.
    Returns
    -------
        AnnData object
    """
    print("Loading points...")
    points = geopandas.read_file(points)

    # Load masks
    print("Loading masks...")
    mask_paths = {"cell": cell, **other}
    masks = pd.Series(mask_paths).apply(_load_masks)

    # Index points for all masks
    print("Indexing points...")
    point_index = masks.apply(lambda mask: _index_points(points[["geometry"]], mask)).T

    # Index masks to cell
    print("Indexing masks...")
    mask_geoms = _index_masks(masks)

    # Main long dataframe for reformatting
    uns_points = pd.concat(
        [
            points[["x", "y", "gene"]].reset_index(drop=True),
            point_index.reset_index(drop=True),
        ],
        axis=1,
    )

    # Remove extracellular points
    uns_points = uns_points.loc[uns_points["cell"] != "-1"]

    # Aggregate points to counts
    print("Formatting AnnData object...")
    if "nucleus" not in uns_points.columns:
        expression = (
            uns_points[["cell", "gene"]]
            .groupby(["cell", "gene"])
            .apply(lambda x: x.shape[0])
            .to_frame()
            .reset_index()
        )
    else:
        # Use nuclear inclusion for splicing/unspliced layers
        expression = (
            uns_points[["cell", "gene", "nucleus"]]
            .groupby(["cell", "gene", "nucleus"])
            .apply(lambda x: x.shape[0])
            .to_frame()
            .reset_index()
        )

    # Create cell x gene matrix
    print("Processing expression...")
    cellxgene = expression.pivot_table(
        index=["cell"], columns=["gene"], aggfunc="sum"
    ).fillna(0)
    cellxgene.columns = cellxgene.columns.get_level_values("gene")

    # Add splice data
    if "nucleus" in uns_points.columns:
        print("Processing splicing...")
        spliced, unspliced = _to_spliced_expression(expression)

    print("Processing point coordinates...")

    # Create scanpy anndata object
    adata = AnnData(X=cellxgene)
    mask_geoms = mask_geoms.reindex(adata.obs.index)
    adata.obs = pd.concat([adata.obs, mask_geoms], axis=1)
    adata.obs.index = adata.obs.index.astype(str)

    # Save spliced/unspliced counts to layers
    if "nucleus" in uns_points.columns:
        adata.layers["spliced"] = spliced
        adata.layers["unspliced"] = unspliced

    # Save cell, gene, and nucleus as categorical type to save memory
    uns_points["cell"] = uns_points["cell"].astype("category")
    uns_points["gene"] = uns_points["gene"].astype("category")

    if "nucleus" in uns_points.columns:
        uns_points["nucleus"] = uns_points["nucleus"].astype("category")

    adata.uns = {"points": uns_points}

    print("Done.")
    return adata


def _load_masks(path):
    """Load GeoDataFrame from path.

    Parameters
    ----------
    path : str
        Path to .shp file.

    Returns
    -------
    GeoDataFrame
        Contains masks as Polygons.
    """
    mask = geopandas.read_file(path)
    mask.index = mask.index.astype(str)

    for i, poly in enumerate(mask["geometry"]):
        if type(poly) == geometry.MultiPolygon:
            print(f"Object at index={i} is a MultiPolygon.")
            print(poly)
            return

    # Cleanup polygons
    # mask.geometry = mask.geometry.buffer(2).buffer(-2)
    # mask.geometry = mask.geometry.apply(unary_union)

    return mask


def _index_masks(masks):
    """Spatially index other masks to cell mask.

    Parameters
    ----------
    masks : dict
        Dictionary of mask GeoDataFrames.

    Returns
    -------
    GeoDataFrame
        [description]
    """
    cell_mask = masks["cell"]

    shapes = cell_mask.copy()
    if "FID" in shapes.columns:
        shapes = shapes.drop("FID", axis=1)

    shapes = shapes.rename(columns={"geometry": "cell_shape"})

    for m, mask in masks.items():
        if m != "cell":
            geometry = (
                geopandas.sjoin(
                    mask, cell_mask, how="left", op="within", rsuffix="cell"
                )
                .dropna()
                .drop_duplicates("index_cell")
                .set_index("index_cell")
                .reindex(cell_mask.index)["geometry"]
            )
            geometry.name = f"{m}_shape"
            shapes[f"{m}_shape"] = geometry

    return shapes


def _index_points(points, mask):
    """Index points to each mask item and save. Assumes non-overlapping masks.

    Parameters
    ----------
    points : GeoDataFrame
        Point coordinates.
    mask : GeoDataFrame
        Mask polygons.
    Returns
    -------
    Series
        Return list of mask indices corresponding to each point.
    """
    index = geopandas.sjoin(points.reset_index(), mask, how="left", op="intersects")

    # remove multiple cells assigned to same point
    index = index.drop_duplicates(subset="index", keep="first")
    index = index.sort_index()
    index = index.reset_index()["index_right"]
    index = index.fillna(-1).astype(str)

    return pd.Series(index)


def concatenate(adatas):
    uns_points = []
    for i, adata in enumerate(adatas):
        points = adata.uns["points"].copy()
        points["cell"] = points["cell"].astype(str) + f"-{i}"

        points["batch"] = i
        uns_points.append(points)

    new_adata = adatas[0].concatenate(adatas[1:])

    uns_points = pd.concat(uns_points)
    points["cell"] = points["cell"].astype("category")
    points["batch"] = points["batch"].astype("category")

    new_adata.uns["points"] = uns_points

    return new_adata


def _to_spliced_expression(expression):
    cell_nucleus = expression.pivot(index=["cell", "nucleus"], columns="gene")
    unspliced = []
    spliced = []
    idx = pd.IndexSlice
    spliced_index = "-1"

    def to_splice_layers(cell_df):
        unspliced_index = (
            cell_df.index.get_level_values("nucleus")
            .drop(spliced_index, errors="ignore")
            .tolist()
        )

        unspliced.append(
            cell_df.loc[idx[:, unspliced_index], :]
            .sum()
            .to_frame()
            .T.reset_index(drop=True)
        )

        # Extract spliced counts for this gene if there are any.
        if spliced_index in cell_df.index.get_level_values("nucleus"):
            spliced.append(
                cell_df.xs(spliced_index, level="nucleus").reset_index(drop=True)
            )
        else:
            # Initialize empty zeros
            spliced.append(
                pd.DataFrame(np.zeros((1, cell_df.shape[1])), columns=cell_df.columns)
            )

    cell_nucleus.groupby("cell").apply(to_splice_layers)

    cells = cell_nucleus.index.get_level_values("cell").unique()

    unspliced = pd.concat(unspliced)
    unspliced.index = cells
    spliced = pd.concat(spliced)
    spliced.index = cells

    spliced = spliced.fillna(0)
    unspliced = unspliced.fillna(0)

    return spliced, unspliced


def to_scanpy(data):
    # Extract points
    expression = pd.DataFrame(
        data.X, index=pd.MultiIndex.from_frame(data.obs[["cell", "gene"]])
    )

    # Aggregate points to counts
    expression = (
        data.obs[["cell", "gene"]]
        .groupby(["cell", "gene"])
        .apply(lambda x: x.shape[0])
        .to_frame()
    )
    expression = expression.reset_index()

    # Remove extracellular points
    expression = expression.loc[expression["cell"] != "-1"]

    # Format as dense cell x gene counts matrix
    expression = expression.pivot(index="cell", columns="gene").fillna(0)
    expression.columns = expression.columns.droplevel(0)
    expression.columns = expression.columns.str.upper()

    # Create scanpy anndata object to use scoring function
    sc_data = sc.AnnData(expression)

    return sc_data



from ._cell_features import cell_area, cell_aspect_ratio
from ._features import (coloc_cluster_genes, coloc_sim, gene_leiden,
                        rasterize_cells)
from ._locfish_features import distance_features, ripley_features, morph_enrichment, nuclear_fraction, moment_stats
from ._spots import detect_spots, distr_to_var, spots_diff, PATTERN_NAMES



from ast import Try
import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)

import os

import dask.dataframe as dd
import geopandas as gpd
import matplotlib.path as mplPath
import numpy as np
import pandas as pd
from astropy.stats import RipleysKEstimator
from joblib import Parallel, delayed
from scipy.spatial import distance, distance_matrix
from scipy.stats import spearmanr
from sklearn.metrics import mean_squared_error
from tqdm.auto import tqdm

from ..preprocessing import get_points

def ripley_features(data, n_cores=1, copy=False):
    """
    Compute 5 features extracted from Ripley L-function.

    https://docs.astropy.org/en/stable/api/astropy.stats.RipleysKEstimator.html#astropy.stats.RipleysKEstimator

    Returns
    -------
    dict : keyed feature values
        max_l : float
        max_gradient : float
        min_gradient : float
        monotony : float
        l_4 : float
    """

    adata = data.copy() if copy else data

    def ripley_per_cell(c, p, cell_name):
        # L/4 where length of the cell L is max dist between 2 points on polygon defining cell border
        c_coords = np.array(c.exterior.coords.xy).T
        l_4_dist = int(distance_matrix(c_coords, c_coords).max() / 4)

        features = []
        for gene, gene_pts in p.groupby("gene"):
            estimator = RipleysKEstimator(
                area=c.area,
                x_min=float(gene_pts["x"].min()),
                x_max=float(gene_pts["x"].max()),
                y_min=float(gene_pts["y"].min()),
                y_max=float(gene_pts["y"].max()),
            )

            # Compute ripley function for r=(1, cell diameter / 2), step size = 1 pixel
            radii = np.linspace(1, l_4_dist * 2, num=l_4_dist * 2)
            stats = estimator.Hfunction(
                data=gene_pts[["x", "y"]].values, radii=radii, mode="none"
            )

            # Max value of the L-function
            max_l = max(stats)

            # Max and min value of the gradient of L
            ripley_smooth = pd.Series(stats).rolling(5).mean()
            ripley_smooth.dropna(inplace=True)

            ripley_gradient = np.gradient(ripley_smooth)
            max_gradient = ripley_gradient.max()
            min_gradient = ripley_gradient.min()

            # Monotony of L-function in the interval
            l_corr = spearmanr(radii, stats)[0]

            # L-function at L/4 where length of the cell L is max dist between 2 points on polygon defining cell border
            l_4 = estimator.Hfunction(
                data=gene_pts[["x", "y"]].values, radii=[l_4_dist], mode="none"
            )[0]

            features.append(
                [cell_name, gene, max_l, max_gradient, min_gradient, l_corr, l_4]
            )

        return features

    # Parallelize points
    cell_features = Parallel(n_jobs=n_cores)(
        delayed(ripley_per_cell)(
            adata.obs.loc[cell_name, "cell_shape"],
            get_points(adata, cells=cell_name),
            cell_name,
        )
        for cell_name in tqdm(adata.obs_names.tolist())
    )

    cell_features = np.array(cell_features).reshape(-1, 7)

    colnames = [
        "cell",
        "gene",
        "max_l",
        "max_l_gradient",
        "min_l_gradient",
        "l_corr",
        "l_at_l4",
    ]
    cell_features = pd.DataFrame(cell_features, columns=colnames)

    for f in colnames[2:]:
        adata.layers[f] = (
            cell_features.pivot(index="cell", columns="gene", values=f)
            .reindex(index=adata.obs_names, columns=adata.var_names)
            .astype(float)
        )

    return adata if copy else None


def distance_features(data, n_cores=1, copy=False):
    """Return 5th, 10th, 20th, and 50th quantiles for distance to cell and nucleus centroids.

    Parameters
    ----------
    data : AnnData
        [description]
    cell : [type]
        [description]

    Returns
    -------
    pd.Series
    """
    adata = data.copy() if copy else data

    def distance_per_cell(c, n, p, cell_name):

        features = []
        for gene, gene_pts in p.groupby("gene"):
            # Calculate normalized distances from points to centroids
            xy = gene_pts[["x", "y"]].values
            cell_centroid_dist = norm_dist_to_centroid(xy, c)
            nucleus_centroid_dist = norm_dist_to_centroid(xy, n)

            # Calculate normalized distances from points to mask; distance = point to closest point on mask
            cell_mask_dist = norm_dist_to_mask(xy, c)
            nucleus_mask_dist = norm_dist_to_mask(xy, n)

            features_raw = [
                cell_centroid_dist,
                nucleus_centroid_dist,
                cell_mask_dist,
                nucleus_mask_dist,
            ]

            # Get quantiles of each distance distribution
            stats = [cell_name, gene]
            for f in features_raw:
                quantiles = [5, 10, 20, 50]
                stats.extend(np.percentile(f, quantiles))

            # There should be 16 features (18 with cell/gene names)
            features.append(np.array(stats))

        return features

    # Parallelize points
    cell_features = Parallel(n_jobs=n_cores)(
        delayed(distance_per_cell)(
            adata.obs.loc[cell_name, "cell_shape"],
            adata.obs.loc[cell_name, "nucleus_shape"],
            get_points(adata, cells=cell_name),
            cell_name,
        )
        for cell_name in tqdm(adata.obs_names.tolist())
    )

    cell_features = np.array(cell_features).reshape(-1, 18)

    # Enuemerate feature names
    feature_labels = ["cell", "gene"]
    features_label_prefix = [
        "cell_centroid_dist",
        "nucleus_centroid_dist",
        "cell_mask_dist",
        "nucleus_mask_dist",
    ]

    for label in features_label_prefix:
        quantiles = [5, 10, 20, 50]
        feature_labels.extend([f"{label}_q{q}" for q in quantiles])

    cell_features = pd.DataFrame(cell_features, columns=feature_labels)

    for f in feature_labels[2:]:
        adata.layers[f] = (
            cell_features.pivot(index="cell", columns="gene", values=f)
            .reindex(index=adata.obs_names, columns=adata.var_names)
            .astype(float)
        )

    return adata if copy else None


def norm_dist_to_centroid(points, mask):
    """Calculate normalized distance between every point in points to centroid of mask.

    Parameters
    ----------
    points : 2d array
        Array of 2D coordinates.
    mask : Polygon
        [description]
    """
    mask_points = np.array(mask.exterior.coords.xy).T
    centroid = np.array(mask.centroid).reshape(1, 2)

    expected_distance = abs(distance.cdist(mask_points, centroid)).mean()
    observed_distance = abs(distance.cdist(points, centroid))
    return observed_distance / expected_distance


def norm_dist_to_mask(points, mask):
    """Calculate normalized distance between every point in points to mask.

    Parameters
    ----------
    points : 2d array
        Array of 2D coordinates.
    mask : Polygon
        [description]
    """
    mask_points = np.array(mask.exterior.coords.xy).T
    centroid = np.array(mask.centroid).reshape(1, 2)

    expected_distance = abs(distance.cdist(mask_points, centroid)).mean()
    observed_distance = abs(distance.cdist(points, mask_points)).min(axis=0)
    return observed_distance / expected_distance


def morph_enrichment(data, n_cores=1, copy=False):
    adata = data.copy() if copy else data

    proportions = [0.05, 0.1]

    def enrichment_per_cell(c, p, cell_name):

        # Mean cell radius size
        mean_cell_radius = distance.cdist(
            np.array(c.centroid).reshape(1, 2), np.array(c.exterior.xy).T
        ).mean()

        # Define proportion of radius to perform morph opening
        morph_masks = {}
        for prop in proportions:
            # Define proportion of radius; compute morph opening
            d = mean_cell_radius * prop
            morph_masks[prop] = c.buffer(distance=-d).buffer(distance=d)

        features = []

        for gene, gene_pts in p.groupby("gene"):

            # Create GeoDataFrame from points
            pts = gpd.GeoDataFrame(geometry=gpd.points_from_xy(gene_pts.x, gene_pts.y))

            # Count fraction of points in mask
            enrichment = [cell_name, gene]
            for prop in proportions:
                n_points = len(gpd.clip(pts, morph_masks[prop]))
                enrichment.append(np.float(n_points) / len(pts))

            features.append(np.array(enrichment))
        return features

    # Parallelize points
    cell_features = Parallel(n_jobs=n_cores)(
        delayed(enrichment_per_cell)(
            adata.obs.loc[cell_name, "cell_shape"],
            get_points(adata, cells=cell_name),
            cell_name,
        )
        for cell_name in tqdm(adata.obs_names.tolist())
    )

    cell_features = np.array(cell_features).reshape(-1, 4)

    # Enuemerate feature labels
    feature_labels = ["cell", "gene"]
    [feature_labels.append(f"morph_{prop}") for prop in proportions]

    cell_features = pd.DataFrame(cell_features, columns=feature_labels)

    for f in feature_labels[2:]:
        adata.layers[f] = (
            cell_features.pivot(index="cell", columns="gene", values=f)
            .reindex(index=adata.obs_names, columns=adata.var_names)
            .astype(float)
        )

    return adata if copy else None


def nuclear_fraction(data, n_cores=1, copy=False):
    """
    Parameters
    ----------
    n_cores : int
        Parameter kept for uniformity. Increasing number of cores will not do anything.
    """
    adata = data.copy() if copy else data

    adata.layers["nuclear_fraction"] = adata.layers["unspliced"] / adata.X

    return adata if copy else None


def moment_stats(data, n_cores=1, copy=False):
    """Calculate polarization index, dispersion index, and peripheral index. Requires both cell and nucleus."""
    adata = data.copy() if copy else data

    def moments_per_cell(c, n, p, cell_name):
        # Calculate of cell-dependent only values
        cell_centroid = np.array(c.centroid).reshape(1, 2)
        grid_points = _poly2grid(c)
        cell_centroid_array = np.tile(cell_centroid, len(grid_points))
        cell_centroid_array = cell_centroid_array.reshape(-1, 2)
        radius_of_gyration = mean_squared_error(grid_points, cell_centroid_array)

        nucleus_centroid = np.array(n.centroid).reshape(1, 2)

        nuclear_moment = _second_moment(nucleus_centroid, grid_points)

        features = []

        for gene, gene_pts in p.groupby("gene"):

            # Calculate distance between point and cell centroids
            pts = gene_pts[["x", "y"]].values
            point_centroid = pts.mean(axis=0).reshape(1, 2)
            offset = distance.cdist(point_centroid, cell_centroid)[0][0]

            # Calculate polarization index
            polarization_index = offset / radius_of_gyration

            # Calculate dispersion index; d-index = 2nd moment of spots, norm. by 2nd moment of uniform spots; moments relative to spot centroid
            dispersion_index = _second_moment(point_centroid, pts) / _second_moment(
                point_centroid, grid_points
            )

            # Calculate peripheral distribution index; d-index = 2nd moment of spots, norm. by 2nd moment of uniform spots; moments relative to nucleus centroid
            peripheral_distribution_index = (
                _second_moment(nucleus_centroid, pts) / nuclear_moment
            )

            features.append(
                np.array(
                    [
                        cell_name,
                        gene,
                        polarization_index,
                        dispersion_index,
                        peripheral_distribution_index,
                    ]
                )
            )

        return features

    # Parallelize points
    cell_features = Parallel(n_jobs=n_cores)(
        delayed(moments_per_cell)(
            adata.obs.loc[cell_name, "cell_shape"],
            adata.obs.loc[cell_name, "nucleus_shape"],
            get_points(adata, cells=cell_name),
            cell_name,
        )
        for cell_name in tqdm(adata.obs_names.tolist())
    )

    cell_features = np.array(cell_features).reshape(-1, 5)

    # Enuemerate feature labels
    feature_labels = [
        "cell",
        "gene",
        "polar_moment",
        "dispersion_moment",
        "peripheral_moment",
    ]

    cell_features = pd.DataFrame(cell_features, columns=feature_labels)

    for f in feature_labels[2:]:
        adata.layers[f] = (
            cell_features.pivot(index="cell", columns="gene", values=f)
            .reindex(index=adata.obs_names, columns=adata.var_names)
            .astype(float)
        )

    return adata if copy else None


# Calculate second moment of points with centroid as reference
def _second_moment(centroid, pts):
    """
    centroid : [1 x 2] float
    pts : [n x 2] float
    """
    radii = distance.cdist(centroid, pts)
    second_moment = np.sum(radii * radii / len(pts))
    return second_moment


def _poly2grid(polygon):
    minx, miny, maxx, maxy = polygon.bounds
    x, y = np.meshgrid(
        np.arange(minx, maxx, step=np.float(20)),
        np.arange(miny, maxy, step=np.float(20)),
    )
    x = x.flatten()
    y = y.flatten()
    xy = np.array([x, y]).T
    polygon_path = mplPath.Path(np.array(polygon.exterior.xy).T)
    polygon_cell_mask = polygon_path.contains_points(xy)
    xy = xy[polygon_cell_mask]

    return xy


def _ripley(points, mask, radii=None):
    """
    Calculate estimation for Ripley H function (0-centered L function).

    Parameters
    ----------
    points : 2d array
    mask : shapely.Polygon
    radii : float list

    Returns
    -------
    list
        ripley l function
    radii
    """
    estimator = RipleysKEstimator(
        area=mask.area,
        x_min=float(points[:, 0].min()),
        x_max=float(points[:, 0].max()),
        y_min=float(points[:, 1].min()),
        y_max=float(points[:, 1].max()),
    )

    # https://docs.astropy.org/en/stable/api/astropy.stats.RipleysKEstimator.html#astropy.stats.RipleysKEstimator
    # if radii is None:
    #     radii = np.linspace(1, np.sqrt(mask.area / 2) ** 0.5, 200)
    ripley = estimator.Hfunction(data=points, radii=radii, mode="none")

    return ripley, radii



from shapely.geometry import Point

def cell_aspect_ratio(data, copy=False):

    adata = data.copy() if copy else data

    def _aspect_ratio(poly):
        # get coordinates of min bounding box vertices around polygon
        x, y = poly.minimum_rotated_rectangle.exterior.coords.xy

        # get length of bound box sides
        edge_length = (Point(x[0], y[0]).distance(Point(x[1], y[1])), Point(x[1], y[1]).distance(Point(x[2], y[2])))

        # length = longest side, width = shortest side
        length, width = max(edge_length), min(edge_length)

        # return long / short ratio
        return length / width

    ar = adata.obs['cell_shape'].apply(lambda poly: _aspect_ratio(poly))
    adata.obs['aspect-ratio'] = ar

    return adata if copy else None


def cell_area(data, copy=False):
    adata = data.copy() if copy else data

    # Calculate pixel-wise area 
    # TODO: unit scale?
    area = adata.obs['cell_shape'].area
    adata.obs['area'] = area

    return adata if copy else None



import warnings
warnings.filterwarnings("ignore")

from functools import partial

import datashader as ds
import datashader.transfer_functions as tf
import geopandas
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import proplot as plot
import seaborn as sns
from datashader.mpl_ext import dsshow
from matplotlib.colors import ListedColormap

from ..preprocessing import get_points
from ..tools import PATTERN_NAMES

matplotlib.rcParams["figure.facecolor"] = (0, 0, 0, 0)

# Masala color palette by Noor
# Note: tested colorblind friendliness, did not do so well
masala_palette = sns.color_palette(
    [
        (1.0, 0.78, 0.27),
        (1.0, 0.33, 0.22),
        (0.95, 0.6, 0.71),
        (0, 0.78, 0.86),
        (0.39, 0.81, 0.63),
        (0.25, 0.25, 0.25),
    ]
)


def spots_diff(data, groups, adjusted=True, ymax=None):
    dl_results = data.uns[f"diff_{groups}"]

    if adjusted:
        rank = "-log10padj"
    else:
        rank = "-log10p"

    g = sns.relplot(
        data=dl_results.sort_values("pattern"),
        x="dy/dx",
        y=rank,
        hue="pattern",
        size=rank,
        sizes=(1, 20),
        col="phenotype",
        height=4,
        linewidth=0,
        palette="tab10",
        facet_kws=dict(xlim=(-1, 1)),
    )

    g.map(plt.axhline, y=-np.log10(0.05), color="gray", lw=1, linestyle="--", zorder=0)
    (
        g.map(plt.axvline, x=0, color="black", lw=1, linestyle="-", zorder=0)
        .set_axis_labels("Marginal Effect (dy/dx)", f"Significance ({rank})")
        .set_titles("{col_name}")
    )
    # g.map(sns.despine, top=False, right=False, bottom=False, left=False)


def gene_umap(data, hue=None, **kwargs):

    umap = data.varm["loc_umap"]
    if hue is not None:
        hue_vector = data.var.loc[umap.index, hue]
    else:
        hue_vector = hue
    ax = sns.scatterplot(data=umap, x=0, y=1, hue=hue_vector, **kwargs)
    ax.legend(bbox_to_anchor=(1.01, 1), loc=2, borderaxespad=0)
    plt.tight_layout()
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_title(hue)
    return ax


def spots_all_distr(data, groupby=None, layer="pattern", stacked=True, legend=True):

    if groupby:
        pattern_distr = (
            data.to_df(layer)
            .groupby(data.obs[groupby])
            .apply(lambda df: pd.Series(df.values.flatten()).value_counts())
        )
        pattern_distr = (pattern_distr / pattern_distr.sum(level=0)).reset_index()

    else:
        groupby = "Sample"
        pattern_distr = pd.Series(data.to_df(layer).values.flatten()).value_counts()
        pattern_distr = (
            (pattern_distr / pattern_distr.sum()).reset_index().reset_index()
        )
        pattern_distr.iloc[:, 0] = "Sample"

    pattern_distr.columns = [groupby, layer, "value"]
    # pattern_distr["value"] *= 100
    pattern_distr = pattern_distr.pivot(index=groupby, columns=layer, values="value")
    pattern_distr = pattern_distr.reindex(columns=PATTERN_NAMES, fill_value=0)
    # pattern_distr.drop("none", axis=1, inplace=True)

    with sns.axes_style(style="white"):
        ax = pattern_distr.plot(
            kind="bar",
            stacked=stacked,
            colormap=ListedColormap(sns.color_palette("muted6", as_cmap=True)),
            width=0.8,
            lw=0,
            figsize=(max(2, pattern_distr.shape[0] / 2), 4),
            # xlim=(0, 100),
            legend=False,
        )

        if pattern_distr.shape[0] == 1:
            ax.set_xlabel("")
        ax.set_ylabel("Total Fraction")
        if legend:
            ax.legend(bbox_to_anchor=(1, 1))
        sns.despine()
        plt.tight_layout()

    return ax


def spots_distr(data, level="cells", sharey=True, binwidth=10, layer="pattern"):
    """Plot localization pattern distributions across all cells (default) or genes.

    Parameters
    ----------
    data : [type]
        [description]
    """
    if level == "cells":
        axis = 1
    elif level == "genes":
        axis = 0

    # Count pattern occurences within axis
    pattern_distr = data.to_df("pattern").apply(
        lambda fiber: fiber.value_counts(), axis=axis
    )

    if axis == 0:
        pattern_distr = pattern_distr.T

    pattern_distr = ((pattern_distr.T * 100) / pattern_distr.sum(axis=1)).T

    distr_long = pattern_distr.drop("none", axis=1)
    distr_long["group"] = None
    distr_long = distr_long.melt(id_vars="group")
    distr_long.columns = ["group", "pattern", "value"]

    palette = sns.color_palette("muted6")
    with sns.axes_style(style="white", rc={"axes.facecolor": (0, 0, 0, 0)}):

        pattern_names = distr_long["pattern"].unique()

        g = sns.FacetGrid(
            distr_long,
            col="pattern",
            hue="pattern",
            col_wrap=3,
            sharey=sharey,
            sharex=False,
            aspect=1.5,
            height=2,
            palette=palette,
            xlim=(0, 100),
        )
        # subplot_kws=dict(facecolor=(0,0,0,0))
        # then we add the densities kdeplots for each pattern
        g.map(
            sns.histplot,
            "value",
            stat="count",
            binwidth=binwidth,
            clip_on=False,
            alpha=1,
        )

        # here we add a horizontal line for each plot
        g.map(plt.axhline, y=0, lw=2, clip_on=False)

        if level == "cells":
            x_unit = "genes"
        else:
            x_unit = "cells"

        # we loop over the FacetGrid figure axes (g.axes.flat) and add the pattern as text with the
        # right color notice how ax.lines[-1].get_color() enables you to access the last line's
        # color in each matplotlib.Axes
        for i, ax in enumerate(g.axes.flat):
            ax.set_title(
                pattern_names[i],
                # fontweight="bold",
                fontsize=15,
                color=ax.lines[-1].get_color(),
            )
            ax.set_xlabel(f"Percent of {x_unit}", fontsize=15)

        # we use matplotlib.Figure.subplots_adjust() function to get the subplots to overlap
        g.fig.subplots_adjust(hspace=0.4)

        # eventually we remove axes titles, yticks and spines
        g.set_titles("")
        plt.setp(ax.get_xticklabels(), fontsize=15)

        g.fig.suptitle(
            f"Localization pattern distribution across {level}",
            ha="center",
            fontsize=18,
            fontweight=20,
        )

        plt.tight_layout()
    return g


def plot_cells(
    data,
    kind="scatter",
    hue=None,
    tile=False,
    cells=None,
    genes=None,
    pattern=None,
    markersize=3,
    alpha=1,
    cmap="blues",
    ncols=4,
    masks="all",
    binwidth=3,
    spread=False,
    legend=False,
    cbar=True,
    frameon=True,
    size=4,
):
    """
    Visualize distribution of variable in spatial coordinates.
    Parameters
    ----------
    data : dict
    type : str
        Options include 'points' and 'heatmap'
    cells: None, list
        Select specified list of cells. Default value of None selects all cells.
    genes: None, list
        Select specified list of genes. Default value of None selects all genes.
    draw_masks: str, list
       masks to draw outlines for.
    """

    # Format cells input
    if cells is None:
        cells = data.obs.index.unique().tolist()
    else:
        cells = list(cells)

    if "-1" in cells:
        warnings.warn(
            "Detected points outside of cells. TODO write clean fx that drops these points"
        )

    # Format genes input
    if genes is None:
        genes = data.var_names
    else:
        # print('Subsetting genes...')
        if type(genes) != list:
            genes = [genes]

        genes = list(set(genes))

    # Add all masks if 'all'
    if masks == "all":
        masks = data.obs.columns[data.obs.columns.str.endswith("shape")]
    else:
        masks = [f"{m}_shape" for m in masks]

    # Convert draw_masks to list
    masks = [masks] if type(masks) is str else masks

    # Subset adata info by genes and cells
    data = data[cells, genes]
    points = get_points(data, cells=cells, genes=genes)

    if pattern:
        points = points.loc[points["pattern"] == pattern]

    points = geopandas.GeoDataFrame(
        points, geometry=geopandas.points_from_xy(points["x"], points["y"])
    )

    # Get masks and points
    shapes = geopandas.GeoDataFrame(data.obs[masks], geometry="cell_shape")
    masks = shapes.columns.tolist()

    # Plot each cell in separate subplots
    if tile:
        ncols = min(ncols, len(shapes))
        nrows = max(1, int(np.ceil(len(shapes) / ncols)))

        # Determine fixed radius of each subplot
        cell_bounds = shapes.bounds
        cell_maxw = (cell_bounds["maxx"] - cell_bounds["minx"]).max()
        cell_maxh = (cell_bounds["maxy"] - cell_bounds["miny"]).max()
        ax_radius = 1.1 * (max(cell_maxw, cell_maxh) / 2)

        # Create subplots
        fig, axs = plot.subplots(
            nrows=nrows, ncols=ncols, sharex=False, sharey=False, axwidth=size, space=0
        )
        axs.format(xticks=[], yticks=[])
        axs.axis(frameon)

        # Plot each cell separately
        for i, ax in enumerate(axs):
            try:
                s = shapes.iloc[[i]]
                p = points.loc[points["cell"] == s.index[0]]

                if i == 0:
                    legend = legend
                else:
                    legend = False
                _plot_cells(
                    masks,
                    s,
                    ax,
                    kind,
                    p,
                    markersize,
                    alpha,
                    binwidth,
                    spread,
                    hue,
                    cmap,
                    size,
                    legend,
                    cbar,
                )

                s_bound = s.bounds
                centerx = np.mean([s_bound["minx"], s_bound["maxx"]])
                centery = np.mean([s_bound["miny"], s_bound["maxy"]])
                ax.set_xlim(centerx - ax_radius, centerx + ax_radius)
                ax.set_ylim(centery - ax_radius, centery + ax_radius)

            except (IndexError, ValueError):
                ax.remove()

    else:
        fig, ax = plt.subplots(1, 1, figsize=(size, size))
        plt.setp(ax, xticks=[], yticks=[])
        ax.axis(frameon)

        _plot_cells(
            masks,
            shapes,
            ax,
            kind,
            points,
            markersize,
            alpha,
            binwidth,
            spread,
            hue,
            cmap,
            size,
            legend,
            cbar,
        )

    if pattern:
        ax.set_title(pattern)

    # plt.close(fig)
    return fig


def _plot_cells(
    masks,
    shapes,
    ax,
    kind,
    points_c,
    markersize,
    alpha,
    binwidth,
    spread,
    hue,
    cmap,
    size,
    legend,
    cbar,
):
    # Plot mask outlines
    for mask in masks:
        shapes.set_geometry(mask).plot(
            color=(0, 0, 0, 0.05), edgecolor=(0, 0, 0, 0.3), ax=ax
        )

    if points_c.shape[0] == 0:
        return ax

    if hue == "pattern" and "none" in points_c[hue].cat.categories:
        points_c[hue].cat.remove_categories("none", inplace=True)
        cmap = dict(
            zip(
                [
                    "cell_edge",
                    "foci",
                    "nuclear_edge",
                    "perinuclear",
                    "protrusions",
                    "random",
                ],
                sns.color_palette("muted6", n_colors=6).as_hex(),
            )
        )

    # Plot points
    if kind == "scatter":
        if hue is None:
            points_c.plot(
                column=hue,
                markersize=markersize,
                color=cmap,
                alpha=alpha,
                ax=ax,
            )
        else:
            points_c.plot(
                column=hue,
                markersize=markersize,
                cmap=cmap,
                alpha=alpha,
                ax=ax,
            )

    # Plot density
    elif kind == "hist":

        if hue:
            aggregator = ds.count_cat(hue)
        else:
            aggregator = ds.count()

        if spread:
            spread = partial(tf.dynspread, threshold=0.5)
        else:
            spread = None

        # Percent of window size
        scaled_binwidth = size * binwidth * 0.5

        artist = dsshow(
            points_c,
            ds.Point("x", "y"),
            aggregator,
            norm="linear",
            cmap=cmap,
            color_key=cmap,
            width_scale=1 / scaled_binwidth,
            height_scale=1 / scaled_binwidth,
            vmin=0,
            shade_hook=spread,
            ax=ax,
        )

        if legend:
            plt.legend(handles=artist.get_legend_elements())

        from mpl_toolkits.axes_grid1 import make_axes_locatable

        if cbar:
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="4%", pad=0.05)
            ax.figure.colorbar(artist, cax=cax, orientation="vertical")
        # plt.tight_layout()

    bounds = shapes.total_bounds
    ax.set_xlim(bounds[0], bounds[2])
    ax.set_ylim(bounds[1], bounds[3])

    return ax


def pheno_to_color(pheno, palette):
    """
    Maps list of categorical labels to a color palette.
    Input values are first sorted alphanumerically least to greatest before mapping to colors.
    This ensures consistent colors regardless of input value order.

    Parameters
    ----------
    pheno : pd.Series
        Categorical labels to map
    palette: None, string, or sequence, optional
        Name of palette or None to return current palette.
        If a sequence, input colors are used but possibly cycled and desaturated.
        Taken from sns.color_palette() documentation.

    Returns
    -------
    dict
        Mapping of label to color in RGBA
    tuples
        List of converted colors for each sample, formatted as RGBA tuples.

    """
    if type(palette) is str:
        palette = sns.color_palette(palette)
    else:
        palette = palette

    values = list(set(pheno))
    values.sort()
    palette = sns.color_palette(palette, n_colors=len(values))
    study2color = dict(zip(values, palette))
    sample_colors = [study2color[v] for v in pheno]
    return study2color, sample_colors



def get_points(data, cells=None, genes=None):

    points = data.uns["points"]

    if cells is not None:
        cells = [cells] if type(cells) is str else cells

        points = points.loc[points["cell"].isin(cells)]

    if genes is not None:
        genes = [genes] if type(genes) is str else genes

        points = points.loc[points["gene"].isin(genes)]

    return points


def set_points(data, cells=None, genes=None, copy=False):
    adata = data.copy() if copy else data
    points = get_points(adata, cells, genes)
    adata.uns["points"] = points
    return adata if copy else None


def filter_points(data, min=None, max=None, copy=False):
    """Select samples with at least min_count and at most max_count points.

    Parameters
    ----------
    data : AnnData
        bento loaded AnnData
    min : int, optional
        minimum points needed to keep sample, by default None
    max : int, optional
        maximum points needed to keep sample, by default None
    copy : bool, optional
        True modifies data in place while False returns a copy of the modified data, by default False
    """
    adata = data.copy() if copy else data
    points = get_points(data)

    expr_flat = adata.to_df().reset_index().melt(id_vars='cell')
    
    if min:
        expr_flat = expr_flat.query(f'value >= {min}')
    
    if max:
        expr_flat = expr_flat.query(f'value <= {max}')

    expr_flat = set(tuple(x) for x in expr_flat[['cell', 'gene']].values)

    sample_ids = [tuple(x) for x in points[['cell', 'gene']].values]
    keep = [True if x in expr_flat else False for x in sample_ids]

    points = points.loc[keep]

    # points = points.groupby(['cell', 'gene']).apply(lambda df: df if df.shape[0] >= 5 else None).reset_index(drop=True)
    adata.uns['points'] = points
    return adata if copy else None




def subsample(data, frac=0.2, copy=True):
    adata = data.copy() if copy else data
    points = get_points(data)

    sampled_pts = points.groupby(['cell', 'gene']).sample(frac=frac)
    
    X = sampled_pts[['cell', 'gene']].pivot_table(
        index="cell", columns="gene", aggfunc=len, fill_value=0
    )

    adata.uns['points'] = sampled_pts

    adata = adata[X.index, X.columns]
    adata.X = X

    return adata if copy else None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bento',
 'bento.datasets',
 'bento.io',
 'bento.plotting',
 'bento.preprocessing',
 'bento.tests',
 'bento.tools']

package_data = \
{'': ['*'], 'bento': ['models/spots/five_pattern/*', 'models/spots/pattern/*']}

install_requires = \
['Shapely>=1.7.1,<2.0.0',
 'anndata>=0.7.6,<0.8.0',
 'astropy>=4.3.post1,<5.0',
 'datashader>=0.13.0,<0.14.0',
 'geopandas>=0.9.0,<0.10.0',
 'ipywidgets>=7.6.3,<8.0.0',
 'leidenalg>=0.8.7,<0.9.0',
 'matplotlib>=3.4.2,<4.0.0',
 'numpy>=1.21.1,<2.0.0',
 'pandas>=1.3.1,<2.0.0',
 'proplot>=0.7.0,<0.8.0',
 'pygeos>=0.10.1,<0.11.0',
 'rasterio>=1.2.6,<2.0.0',
 'scanpy>=1.8.1,<2.0.0',
 'scipy>=1.7.0,<2.0.0',
 'seaborn>=0.11.1,<0.12.0',
 'skorch>=0.10.0,<0.11.0',
 'torchvision>=0.10.0,<0.11.0',
 'tqdm>=4.61.2,<5.0.0',
 'umap-learn>=0.5.1,<0.6.0']

setup_kwargs = {
    'name': 'bento-tools',
    'version': '0.1.1',
    'description': 'A toolkit for subcellular analysis of RNA organization',
    'long_description': '# Bento\n\nBento is a toolkit for ingesting, visualizing, and analyzing spatial transcriptomics data at subcellular resolution. \n\n## Installation\n\nUse the package manager [pip](https://pip.pypa.io/en/stable/) to install bento.\n\n```bash\npip install bento-tools\n```\n\n## Usage\n\n```python\nimport bento\n# Todo\n```\n\n## Contributing\nPull requests are welcome. For major changes, please open an issue first to discuss what you would like to change.\n\n## License\n[MIT](https://choosealicense.com/licenses/mit/)',
    'author': 'Clarence Mah',
    'author_email': 'ckmah@ucsd.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



import pickle
import warnings

import numpy as np
import pandas as pd
import statsmodels.formula.api as sfm
import torch
import torch.nn as nn
import torch.nn.functional as F
from joblib import Parallel, delayed
from patsy import PatsyError
from sklearn.preprocessing import OneHotEncoder
from skorch import NeuralNetClassifier
from skorch.callbacks import Checkpoint
from statsmodels.stats.multitest import multipletests
from statsmodels.tools.sm_exceptions import ConvergenceWarning, PerfectSeparationError
from torchvision import datasets, transforms
from tqdm.auto import tqdm

import bento

warnings.simplefilter("ignore", ConvergenceWarning)


PATTERN_NAMES = [
    "cell_edge",
    "foci",
    "nuclear_edge",
    "perinuclear",
    "protrusions",
    "random",
]


def detect_spots(
    patterns, imagedir, batch_size=1024, device="auto", model="pattern", copy=False
):
    """
    Detect and label localization patterns.
    TODO change patterns to be iterable compatible with skorch.predict_proba

    Parameters
    ----------
    patterns : [type]
        [description]
    imagedir : str
        Folder for rasterized images.
    Returns
    -------
    [type]
        [description]
    """
    adata = patterns.copy() if copy else patterns

    # Default to gpu if possible. Otherwise respect specified parameter
    if device == "auto":
        device = "cuda" if torch.cuda.is_available() else "cpu"

    model_dir = "/".join(bento.__file__.split("/")[:-1]) + "/models/spots/" + model

    model_params = pickle.load(open(f"{model_dir}/params.p", "rb"))

    # Load model
    modules = dict(pattern=SpotsModule, five_pattern=FiveSpotsModule)
    module = modules[model](**model_params)

    net = NeuralNetClassifier(module=module, batch_size=batch_size, device=device)
    net.initialize()
    net.load_params(checkpoint=Checkpoint(dirname=f"{model_dir}"))

    dataset = datasets.ImageFolder(
        imagedir,
        transform=transforms.Compose([transforms.Grayscale(), transforms.ToTensor()]),
    )

    # 2d array, sample by n_classes
    pred_prob = net.predict_proba(dataset)

    if isinstance(module, FiveSpotsModule):
        label_names = [
            "cell_edge",
            "nuclear_edge",
            "perinuclear",
            "protrusions",
            "random",
        ]
    else:
        label_names = PATTERN_NAMES

    encoder = OneHotEncoder(handle_unknown="ignore").fit(
        np.array(label_names).reshape(-1, 1)
    )

    # Cell gene names
    sample_names = [
        str(path).split("/")[-1].split(".")[0].rsplit("_", 1)
        for path, _ in dataset.imgs
    ]
    spots_pred_long = pd.DataFrame(sample_names, columns=["cell", "gene"])

    # Pair probabilities to samples
    spots_pred_long[label_names] = pred_prob

    # Flatten p(patterns) for all genes in each cell
    loc_embed = (
        spots_pred_long.pivot(index="cell", columns="gene", values=label_names)
        .fillna(0)
        .reindex(adata.obs_names)
        .values
    )

    # Save flattened probabilities
    adata.obsm[f"{model}_embed"] = loc_embed

    # TODO use spares matrices to avoid slow/big df pivots
    # https://stackoverflow.com/questions/55404617/faster-alternatives-to-pandas-pivot-table
    # Build "pattern" genexcell layer, where values are pattern labels
    spots_pred_long["label"] = encoder.inverse_transform(pred_prob >= 0.5)

    pattern_labels = (
        spots_pred_long.pivot(index="cell", columns="gene", values="label")
        .fillna("none")
        .reindex(index=adata.obs_names, columns=adata.var_names, fill_value="none")
    )

    pattern_labels.columns.name = "gene"

    adata.layers[model] = pattern_labels

    # Annotate points with pattern labels
    plabels_long = pattern_labels.reset_index().melt(id_vars="cell")
    plabels_long = plabels_long.rename({"value": model}, axis=1)

    # Overwrite existing values
    if model in adata.uns["points"].columns:
        adata.uns["points"].drop([model], axis=1, inplace=True)

    # Annotate points
    adata.uns["points"] = adata.uns["points"].merge(
        plabels_long, how="left", on=["cell", "gene"]
    )

    # Save pattern values as categorical to save memory
    adata.uns["points"][model] = adata.uns["points"][model].astype("category")

    # Save to adata.var
    distr_to_var(adata, model)

    return adata if copy else None


def distr_to_var(data, layer, copy=False):
    """Computes frequencies of input layer values across cells and across genes.
    Assumes layer values are categorical.

    Parameters
    ----------
    data : [type]
        [description]
    layer : [type]
        [description]
    copy : bool, optional
        [description], by default False

    Returns
    -------
    [type]
        [description]
    """
    adata = data.copy() if copy else data

    # Save frequencies across genes to adata.var
    gene_summary = (adata.to_df(layer).apply(lambda g: g.value_counts()).fillna(0)).T
    adata.var[gene_summary.columns] = gene_summary

    # Save frequencies across cells to adata.obs
    cell_summary = (
        data.to_df(layer).apply(lambda row: row.value_counts(), axis=1).fillna(0)
    )
    adata.obs[cell_summary.columns] = cell_summary

    return adata if copy else None


def get_conv_dim(in_size, padding, dilation, kernel_size, stride):
    outsize = 1 + (in_size + 2 * padding - dilation * (kernel_size - 1) - 1) / stride
    return int(outsize)


class DataFlatten:
    def fit(self, X, y=None):
        return self

    def transform(self, X):
        X = X.reshape(X.shape[0], -1)
        return X


class DataReshape:
    def fit(self, X, y=None):
        return self

    def transform(self, X):
        X = X.reshape(X.shape[0], 1, 64, 64)
        return X


class SpotsModule(nn.Module):
    def __init__(
        self,
        n_conv_layers,
        in_dim,
        out_channels,
        kernel_size,
        f_units_l0,
        f_units_l1,
    ) -> None:
        super().__init__()
        conv_layers = []

        in_channels = 1
        in_dim = in_dim

        # Stack (convolutions + batchnorm + activation) + maxpool
        for i in range(n_conv_layers):
            conv_layers.append(
                nn.Conv2d(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    kernel_size=kernel_size,
                )
            )
            conv_layers.append(nn.BatchNorm2d(out_channels))
            conv_layers.append(nn.ReLU())

            # Compute convolved output dimensions
            in_dim = get_conv_dim(
                in_dim, padding=0, dilation=1, kernel_size=kernel_size, stride=1
            )

            in_channels = out_channels
            out_channels *= 2

        out_channels = int(out_channels / 2)

        conv_layers.append(nn.MaxPool2d(2, 2))
        in_dim = int(in_dim / 2)

        # We optimize the number of layers, hidden units and dropout ratio in each layer.
        fc_layers = [nn.Flatten()]

        # Compute flatten size
        in_features = out_channels * in_dim * in_dim
        for i in [f_units_l0, f_units_l1]:
            out_features = i
            fc_layers.append(nn.Linear(in_features, out_features))
            fc_layers.append(nn.BatchNorm1d(out_features))
            fc_layers.append(nn.ReLU())

            in_features = out_features

        fc_layers.append(nn.Linear(in_features, 6))

        self.model = nn.Sequential(*[*conv_layers, *fc_layers])

    def forward(self, x):
        x = self.model(x)

        x = F.softmax(x, dim=-1)

        return x


class FiveSpotsModule(nn.Module):
    def __init__(
        self,
        n_conv_layers,
        in_dim,
        out_channels,
        kernel_size,
        f_units_l0,
        f_units_l1,
    ) -> None:
        super().__init__()
        conv_layers = []

        in_channels = 1
        in_dim = in_dim

        # Stack (convolutions + batchnorm + activation) + maxpool
        for i in range(n_conv_layers):
            conv_layers.append(
                nn.Conv2d(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    kernel_size=kernel_size,
                )
            )
            conv_layers.append(nn.BatchNorm2d(out_channels))
            conv_layers.append(nn.ReLU())

            # Compute convolved output dimensions
            in_dim = get_conv_dim(
                in_dim, padding=0, dilation=1, kernel_size=kernel_size, stride=1
            )

            in_channels = out_channels
            out_channels *= 2

        out_channels = int(out_channels / 2)

        conv_layers.append(nn.MaxPool2d(2, 2))
        in_dim = int(in_dim / 2)

        # We optimize the number of layers, hidden units and dropout ratio in each layer.
        fc_layers = [nn.Flatten()]

        # Compute flatten size
        in_features = out_channels * in_dim * in_dim
        for i in [f_units_l0, f_units_l1]:
            out_features = i
            fc_layers.append(nn.Linear(in_features, out_features))
            fc_layers.append(nn.BatchNorm1d(out_features))
            fc_layers.append(nn.ReLU())

            in_features = out_features

        fc_layers.append(nn.Linear(in_features, 5))

        self.model = nn.Sequential(*[*conv_layers, *fc_layers])

    def forward(self, x):
        x = self.model(x)

        x = F.softmax(x, dim=-1)

        return x


def spots_diff(
    data, phenotype=None, continuous=False, combined=False, n_cores=1, copy=False
):
    """Gene-wise test for differential localization across phenotype of interest.

    Parameters
    ----------
    data : AnnData
        Anndata formatted spatial data.
    phenotype : str
        Variable grouping cells for differential analysis. Must be in data.obs_names.
    continuous : bool
        Whether the phenotype is continuous or categorical. By default False.
    n_cores : int, optional
        cores used for multiprocessing, by default 1
    copy : bool, optional
        Return view of AnnData if False, return copy if True. By default False.
    """
    adata = data.copy() if copy else data

    # Parallelize on chunks
    patterns = adata.layers["pattern"].T
    phenotype_vector = adata.obs[phenotype].tolist()

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        diff_output = Parallel(n_jobs=n_cores)(
            delayed(_spots_diff_gene)(
                gene, gp, phenotype, phenotype_vector, continuous, combined
            )
            for gene, gp in tqdm(zip(adata.var_names, patterns), total=len(patterns))
        )

    # Format pattern column
    diff_output = pd.concat(diff_output)

    # FDR correction
    results_adj = []
    for _, df in diff_output.groupby("pattern"):
        df["padj"] = multipletests(df["pvalue"], method="hs")[1]
        results_adj.append(df)

    results_adj = pd.concat(results_adj).dropna()

    # -log10pvalue, padj
    results_adj["-log10p"] = -np.log10(results_adj["pvalue"].astype(np.float32))
    results_adj["-log10padj"] = -np.log10(results_adj["padj"].astype(np.float32))

    # Cap significance values
    results_adj.loc[results_adj["-log10p"] > 20, "-log10p"] = 20
    results_adj.loc[results_adj["-log10padj"] > 12, "-log10padj"] = 12

    # Sort results
    results_adj = results_adj.sort_values("pvalue")

    # Save back to AnnData
    adata.uns[f"diff_{phenotype}"] = results_adj

    return adata if copy else None


def _spots_diff_gene(gene, patterns, phenotype, phenotype_vector, combined):
    """Perform pairwise comparison between groupby and every class.

    Parameters
    ----------
    chunk : tuple

    Returns
    -------
    DataFrame
        Differential localization test results. [# of patterns, ]
    """
    results = []
    # Series denoting pattern frequencies
    pattern_dummies = pd.get_dummies(patterns)
    pattern_dummies = pattern_dummies.drop("none", axis=1)
    pattern_names = pattern_dummies.columns.tolist()

    # One hot encode categories
    group_dummies = pd.get_dummies(pd.Series(phenotype_vector))
    group_dummies.columns = [f"{phenotype}_{g}" for g in group_dummies.columns]
    group_names = group_dummies.columns.tolist()
    group_data = pd.concat([pattern_dummies, group_dummies], axis=1)
    group_data.columns = group_data.columns.astype(str)

    # Perform one group vs rest logistic regression
    for g in group_names:
        try:
            res = sfm.logit(
                formula=f"{g} ~ {' + '.join(pattern_names)}", data=group_data
            ).fit(disp=0)

            # Look at marginal effect of each pattern coefficient
            r = res.get_margeff(dummy=True).summary_frame()
            r["gene"] = gene
            r["phenotype"] = g
            # r["pattern"] = p

            r.columns = [
                "dy/dx",
                "std_err",
                "z",
                "pvalue",
                "ci_low",
                "ci_high",
                "gene",
                "phenotype",
                # "pattern",
            ]
            # r.reset_index(drop=True, inplace=True)
            r = r.reset_index().rename({"index": "pattern"}, axis=1)

            results.append(r)
        except (np.linalg.LinAlgError, PerfectSeparationError, PatsyError):
            continue

    return pd.concat(results) if len(results) > 0 else None
